#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='{{package}}',
    version='1.0.0',
    license='<<insert here>>',
    author='<<your name here>>',
    author_email='you@example.com',
    url='<<your company url>>',
    description="Resilient Circuits Components for '{{package}}'",
    long_description="Resilient Circuits Components for '{{package}}'",
    install_requires=[
        'resilient_circuits>=30.0.0'
    ],
    packages=find_packages(),
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
    ],
    entry_points={
        "resilient.circuits.components": [{%set comma=joiner(",")%}{%for function_name in functions.keys()%}{{comma()}}
            "{{function_name|camel}}FunctionComponent = {{package}}.components.{{function_name}}:FunctionComponent"{%endfor%}
        ],
        "resilient.circuits.configsection": ["gen_config = {{package}}.util.config:config_section_data"],
        "resilient.circuits.customize": ["customize = {{package}}.util.customize:customization_data"],
        "resilient.circuits.selftest": ["selftest = {{package}}.util.selftest:selftest_function"]
    }
)
